#-*- Mode: CPerl -*-

## File: WhoDat::Generator.pm
## Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
## Description:
##  + WhoDat: stochastic generator
##======================================================================

package WhoDat::Generator;
use WhoDat::Object;
use Carp;

our @ISA = qw(WhoDat::Object);


##======================================================================
## Structure
##  + object structure:
##     { counts=>{ ''=>$total_count, $key1=>{''=>$key1_count, $val1=>$count_1_1, ...},
##       state=>$current_key,
##       weight=>$current_weight,
##     }
sub new {
  my $that = shift;
  return $that->SUPER::new(counts=>{''=>0},state=>undef,weight=>0,@_);
}

##======================================================================
## Dummy

## undef = $gen->reset()
##  + resets state (emergency)
sub reset {
  $_[0]{state} = undef;
}


##======================================================================
## Generation

## $next_state = $gen->next()
## $next_state = $gen->next($cur_state)
## $next_state = $gen->next($cur_state,$weight_float)
##   + gets & returns valid value for state $key, updates $gen->{state}
sub next {
  my ($gen,$key,$weight) = @_;
  my ($val);

  while (!defined($val)) {
  ##-- ensure state
    $key = $gen->{state} if (!defined($key));
    if (!defined($key)) {
      my @keys = grep { $_ ne '' } keys (%{$gen->{counts}});
      $key = $keys[rand(@keys)];
    }

    ##-- ensure weight
    $weight = rand(65535.0)/65535.0 if (!defined($weight));
    $gen->{weight} = $weight;

    ##-- sanity check
    my $vals = $gen->{counts}{$key};
    if (!$vals || !%$vals || !$vals->{''} || scalar(keys(%$vals)==1)) {
      $key = undef;
      $gen->reset;
      if (defined($_[1])) {
	carp(ref($gen), "::next(): ignoring invalid key '$key'\n");
      }
      next;
    }

    ##-- get next value
    my $mass = $gen->{mass} = $weight * $vals->{''};
    my ($valct);

    while (!defined($val)) {	##-- rotation hack
      while (($val,$valct)=each(%$vals)) {
	next if ($val eq '');
	$mass -= $valct;
	last if ($mass <= 0);
      }
    }
  }

  ##-- update state
  my @qs = split(/\t+/,$key);
  shift(@qs);
  push(@qs,$val);
  $gen->{state} = join("\t", @qs);

  ##-- return
  return $val;
}


1;

##======================================================================
## Docs
=pod

=head1 NAME

WhoDat::Generator - common generator object for WhoDat

=head1 SYNOPSIS

 use WhoDat::Generator;

 $gen = WhoDat::Generator->new(counts=>{''=>$total,$key1=>\%key1counts,...});
 $gen->{state} = $initial_key;
 $next_state   = $gen->next();

=cut

##======================================================================
## Description
=pod

=head1 DESCRIPTION

WhoDat::Generator is a an abstract base class which implements
a basic state-dependent stochastic selection algorithm.

=cut

##======================================================================
## Methods
=pod

=head1 METHODS

=over 4


=item $gen-E<gt>reset()

Reset to default state.
Overridable by child classes.

=item $gen-E<gt>next()

=item $gen-E<gt>next($current_state)

=item $gen-E<gt>next($current_state,$weight)

Stochastic selection of next state given current state.
Returns the final element of the new selected state, and
updates $gen-E<gt>{state} to contain the full key.

$current_state specifies the source state for selection,
which defaults to $gen-E<gt>{state}.
If specified, $weight should be a floating point number between
0 and 1 (inclusive).

=back

=cut



##======================================================================
## Footer
=pod

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

=head1 AUTHOR

Bryan Jurish E<lt>jurish@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2005, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1),
WhoDat(3perl).

=cut
