#-*- Mode: CPerl -*-

## File: WhoDat::Ngrams.pm
## Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
## Description:
##  + WhoDat: N-gram distribution
##  + really just a placeholder for TnT-style I/O
##======================================================================

package WhoDat::Ngrams;
use WhoDat::Object;
use WhoDat::Generator;
our @ISA = qw(WhoDat::Generator);

##======================================================================
## Constructor
## $ngs = WhoDat::Ngrams->new(%args)
##  + see WhoDat::Generator
##  + new %args:
##  + object structure:
##      nmax => $max_N,
##      ngs => {
##              ''=> $total_count,
##              "$tag_1\t...\t$tag_{N-1}" => {
##                                            ''    => $total,
##                                            $tagN => $count,
##                                            ...,
##                                           },
##              ...
##             }
##      nmax  => $maximum_ngram_length,  ##-- default: 3
sub new {
  my $that = shift;
  return $that->SUPER::new(nmax=>3,
			   counts=>{''=>0},
			   state=>"__\$\t__\$",
			   @_,
			  );
}

##======================================================================
## Overrides
sub reset {
  $_[0]{state} = "__\$\t__\$";
}

##======================================================================
## Utilities
##======================================================================

## @components = $d->split($event)
##  + splits $event into components
sub split {
  return ($_[0]{nmax}
	  ? CORE::split(/\t+/, $_[1], $_[0]{nmax}-1)
	  : CORE::split(/\t+/, $_[1]));
}



##======================================================================
## I/O: Native

## $bool = $obj->saveNativeFh($fh,%args)
##   -- n/a

## $obj = $class_or_obj->loadNativeFh($fh,%args)
sub loadNativeFh {
  my ($ngs,$fh,%args) = @_;
  $ngs = $ngs->new(%args) if (!ref($ngs));

  my ($line,$f);
  my $nz = $ngs->{counts};
  my @ngtmp = qw();
  my @ng = qw();
  my ($key,$val);
  while (defined($line=$fh->getline)) {
    chomp $line;
    next if ($line =~ /^%%/ || $line =~ /^\s*$/);
    $line =~ s/\s+$//;

    @ngtmp  = CORE::split(/\t/, $line);
    $f      = pop(@ngtmp);

    $ng[$_] = $ngtmp[$_] foreach (grep { $ngtmp[$_] ne '' } (0..$#ngtmp));
    splice(@ng, $#ngtmp+1);

    next if (@ng != $ngs->{nmax}); ##-- ignore all but NMAX-grams

    $val = pop(@ng);
    $key = join("\t", @ng);

    $nz->{''}         += $f;
    $nz->{$key}{''}   += $f;
    $nz->{$key}{$val} += $f;

    ##-- hack: handle BOS
    if ($ng[0] eq '__$' && $val ne '__$') {
      $val = pop(@ng);
      push(@ng, '__$');
      $key = join("\t", @ng);
      $nz->{''}         += $f;
      $nz->{$key}{''}   += $f;
      $nz->{$key}{$val} += $f;
      $ng[$#ng] = $val;
    }
  }

  return $ngs;
}


1;

##======================================================================
## Docs
=pod

=head1 NAME

WhoDat::Ngrams - Ngram table for WhoDat

=head1 SYNOPSIS

 use WhoDat::Ngrams;

 $ngs = WhoDat::Ngrams->loadFile("mymodel.123");
 while (($tag=$ngs->next()) ne "__\$") { print $tag, " "; }

=cut

##======================================================================
## Description
=pod

=head1 DESCRIPTION

WhoDat::Ngrams is a top-level class for stochastic sequence generation
via Markov chains.
Data may be loaded from a text n-gram frequency model as described
in the mootfiles(5) manpage.  WhoDat::Ngrams inherits from
WhoDat::Generator and WhoDat::Object.

The depth of the n-gram model (the value of $n) may be specified
at object creation time by the keyword parameter "nmax", which
defaults to 3 (trigram model).

=cut

##======================================================================
## Footer
=pod

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

=head1 AUTHOR

Bryan Jurish E<lt>jurish@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2005, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1),
mootfiles(5),
WhoDat(3perl),
WhoDat::Generator(3perl),
WhoDat::Object(3perl).

=cut
