#-*- Mode: Perl -*-

## File: WhoDat::Object.pm
## Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
## Description:
##  + WhoDat: base class
##======================================================================

package WhoDat::Object;
use IO::File;
use Exporter;
use Carp;

##======================================================================
## Exports
our @ISA = qw(Exporter);
our @EXPORT = qw();
our %EXPORT_TAGS =
  (
   utils  => [qw(dummy)],
  );
$EXPORT_TAGS{all} = [map { @$_ } values(%EXPORT_TAGS)];
our @EXPORT_OK = @{$EXPORT_TAGS{all}};

##======================================================================
## GLOBALS

##======================================================================
## Generic Constructor
## $obj = $class->new(@args)
##  + object structure= hash { %args }
##  + calls $obj->init(@args) on empty hash obj
sub new {
  my $that = shift;
  my $self = bless {}, ref($that)||$that;
  $self->init(@_) if (@_);
  return $self;
}

##======================================================================
## Dummy sub
## \&dummy_sub = dummy($name)
sub dummy {
  my $name = shift;
  return sub { confess( ref($_[0])||$_[0] , "::${name}() not implemented"); };
}

##======================================================================
## Generic initializer
##  + object structure= hash { \%args }
##  + just takes over \%args
sub init {
  my ($obj,%args) = @_;
  @$obj{keys(%args)} = values(%args);
}

##======================================================================
## Deep copy
##  $copy = copy($obj)
#*clone = \&copy;
#sub copy { return Storable::dclone($_[0]); }

sub copy {
  my $src = shift;
  my $ref = ref($src);
  return $src if (!$ref);

  if (UNIVERSAL::isa($ref,'HASH')) {
    my $dst = {};
    my ($k,$v);
    while (($k,$v)=each(%$src)) {
      $dst->{$k} = ref($v) ? copy($v) : $v;
    }
    bless($dst,$ref) if ($ref ne 'HASH');
    return $dst;
  }
  elsif (UNIVERSAL::isa($ref,'ARRAY')) {
    my $dst = [ map { ref($_) ? copy($_) : $_ } @$src ];
    bless($dst,$ref) if ($ref ne 'ARRAY');
  }
  elsif (UNVIERSAL::isa($ref,'SCALAR')) {
    my $dst = ref($$src) ? copy($$src) : $$src;
    my $dstr = \$dst;
    bless($dstr,$ref) if ($ref ne 'SCALAR');
    return $dstr;
  }

  confess(__PACKAGE__, "::copy(): cannot handle object of type \`$ref' -- ignoring!\n");
  return undef;
}


########################################################################
## I/O
########################################################################
package WhoDat::Object;

##======================================================================
## I/O: Native: save
##======================================================================

## $bool = $obj->saveNativeFile($file,@args)
##  + calls saveNativeFh()
sub saveNativeFile {
  my ($obj,$file) = splice(@_,0,2);
  my $fh = ref($file) ? $file : IO::File->new(">$file");
  if (!$fh) {
    confess( __PACKAGE__ , "::saveNativeFile(): open failed for '$file': $!");
    return undef;
  }
  binmode($fh,':utf8');

  my $rc = $obj->saveNativeFh($fh,@_);
  $fh->close() if (!ref($file));
  return $rc;
}

## $bool = $obj->saveNativeFh($fh,@args)
##  + dummy
*saveNativeFh = WhoDat::Object::dummy('saveNativeFh');


##======================================================================
## I/O: Native: load
##======================================================================

## $obj = $obj->loadNativeString($str,@args)
sub loadNativeString {
  my ($obj,$str) = splice(@_,0,2);
  my $fh = IO::Scalar->new(\$str);
  if (!$fh) {
    confess( __PACKAGE__ , "::loadNativeString(): open failed: $!");
    return undef;
  }
  #binmode($fh,':utf8');

  my $rc = $obj->loadNativeFh($fh,@_);
  $fh->close();
  return $rc;
}

## $obj = $obj->loadNativeFile($file,@args)
##  + calls saveNativeFh()
sub loadNativeFile {
  my ($obj,$file) = splice(@_,0,2);
  my $fh = ref($file) ? $file : IO::File->new("<$file");
  if (!$fh) {
    confess( __PACKAGE__ , "::loadNativeFile(): open failed for '$file': $!");
    return undef;
  }
  #binmode($fh,':utf8');

  my $rc = $obj->loadNativeFh($fh,@_);
  $fh->close() if (!ref($file));
  return $rc;
}

## $obj = $obj->loadNativeFh($fh,@args)
##  + calls loadNativeStr()
*loadNativeFh = WhoDat::Object::dummy('loadNativeFh');

##======================================================================
## I/O: Generic: Modes
##======================================================================

## \%ioModes = $class_or_obj->ioModes()
##   + %ioModes = { $modeString => \%modeInfo, ... , 'DEFAULT'=>\%defaultModeInfo }
##   + %modeInfo = { loadFh=>$loadFh_coderef_or_name, saveFh=>$saveFh_coderef_or_name }
##   + subclasses may redefine this, but use of 'registerIOMode()' is reccommended
sub ioModes {
  return {
	  'native'  => __PACKAGE__->ioModeHash('Native'),
	  #'xml'     => __PACKAGE__->ioModeHash('XML'),
	  #'bin'     => __PACKAGE__->ioModeHash('Bin'),
	  ##--
	  'DEFAULT' => __PACKAGE__->ioModeHash('Native'),
	 };

}

## \%modeHash = $class_or_obj->ioModeHash($filenameInfix)
##  + populates a mode hash for $filenameInfix:
##    (load|save)${filenameInfix}(String|File|Fh)
sub ioModeHash {
  my $infix = $_[1];
  return
    {
     loadFh=>"load${infix}Fh",
     saveFh=>"save${infix}Fh",
     loadString=>"load${infix}String",
     saveString=>"save${infix}String",
     loadFile=>"load${infix}File",
     saveFile=>"save${infix}File",
    };
}

## undef = $class_or_obj->registerIOMode($modeString,\%modeInfo)
##  + redefines 'sub ${class}::ioModes() {...}'
sub registerIOMode {
  my ($that,$modeString,$modeInfo) = @_;
  my $modes = {};
  %$modes = %{$that->ioModes()} if (UNIVERSAL::can($that,'ioModes'));
  $modes->{$modeString} = $modeInfo;
  eval {
    no warnings 'redefine';
    *{(ref($that)||$that)."::ioModes"} = sub { return $modes; }
  }
}


## \@fileSuffixModes = $class_or_obj->fileSuffixModes()
##   + @fileSuffixModes = ( {regex=>$fileRegex, mode=>$ioModeStr}, ... )
##   + subclasses *may* redefine this method directly, but use of
##     'registerFileSuffix()' is highly reccommended
sub fileSuffixModes {
  return [
	  #{regex=>qr/\.bin$/i,    mode=>'bin'},
	  #{regex=>qr/\.xml$/i,    mode=>'xml'},
	  {regex=>qr/\.native$/i, mode=>'native'},
	 ];
}

## undef = $class_or_obj->registerFileSuffix($suffix,$modeString)
##  + redefines 'sub ${class}::fileSuffixModes() {...}'
sub registerFileSuffix {
  my ($that,$suffix,$modeString) = @_;
  my $suffixes = [];
  @$suffixes = @{$that->fileSuffixModes()} if (UNIVERSAL::can($that,'fileSuffixModes'));
  unshift(@$suffixes, {regex=>qr/${suffix}$/i, mode=>$modeString});
  eval {
    no warnings 'redefine';
    *{(ref($that)||$that)."::fileSuffixModes"} = sub { return $suffixes; }
  }
}


##======================================================================
## I/O: Generic: Load
##======================================================================


## $obj = $obj->loadGenericFile($filename,@args)
*loadFile = \&loadGenericFile;
sub loadGenericFile {
  my ($that,$filename,%args) = @_;
  my $mode = $args{mode};
  delete($args{mode});

  ##-- guess mode from filename suffix
  if (!defined($mode)) {
    my $suffs = $that->fileSuffixModes();
    foreach $suffSpec (@$suffs) {
      if ($filename =~ $suffSpec->{regex}) {
	$mode = $suffSpec->{mode};
	last;
      }
    }
  }
  $mode = 'DEFAULT' if (!defined($mode));

  ##-- first try file method directly
  if (defined($sub=$that->ioModes->{$mode}{loadFile})
      &&
      (ref($sub) || defined($sub=UNIVERSAL::can($that,$sub))))
    {
      return $sub->($that,$filename,%args);
    }

  ##-- get load sub: fallback to fh method
  if (!defined($sub=$that->ioModes->{$mode}{loadFh})
      ||
      (!ref($sub) && !defined($sub=UNIVERSAL::can($that,$sub))))
    {
      confess((ref($that)||$that), "::loadGenericFile(): could not find load sub for mode '$mode': $!");
      return undef;
    }

  ##-- get fh
  my $fh = IO::File->new("<$filename");
  if (!$fh) {
    confess((ref($that)||$that) , "::loadGenericFile(): open failed for '$filename': $!");
    return undef;
  }
  binmode($fh,':utf8');

  my $rc = $sub->($that, $fh, %args);
  $fh->close();
  return $rc;
}


## $obj = $obj->loadGenericFh($fh,%args)
*loadFh = \&loadGenericFh;
sub loadGenericFh {
  my ($that,$fh,%args) = @_;
  my $mode = defined($args{mode}) ? $args{mode} : 'DEFAULT';
  delete($args{mode});

  ##-- get load sub: fh method only
  if (!defined($sub=$that->ioModes->{$mode}{loadFh})
      ||
      (!ref($sub) && !defined($sub=UNIVERSAL::can($that,$sub))))
    {
      confess((ref($that)||$that), "::loadGenericFh(): could not find load sub for mode '$mode': $!");
      return undef;
    }

  my $rc = $sub->($that, $fh, %args);
  $fh->close();
  return $rc;
}




##======================================================================
## I/O: Generic: Save
##======================================================================

## $bool = $obj->saveGenericFile($filename,%args)
*saveFile = \&saveGenericFile;
sub saveGenericFile {
  my ($that,$filename,%args) = @_;
  my $mode = $args{mode};
  delete($args{mode});

  ##-- guess mode from filename suffix
  if (!defined($mode)) {
    my $suffs = $that->fileSuffixModes();
    foreach $suffSpec (@$suffs) {
      if ($filename =~ $suffSpec->{regex}) {
	$mode = $suffSpec->{mode};
	last;
      }
    }
  }
  $mode = 'DEFAULT' if (!defined($mode));

  ##-- get save sub: first try file method directly
  if (defined($sub=$that->ioModes->{$mode}{saveFile})
      &&
      (ref($sub) || defined($sub=UNIVERSAL::can($that,$sub))))
    {
      return $sub->($that,$filename,%args);
    }


  ##-- get save sub: fallback to fh method
  if (!defined($sub=$that->ioModes->{$mode}{saveFh})
      ||
      (!ref($sub) && !defined($sub=UNIVERSAL::can($that,$sub))))
    {
      confess((ref($that)||$that), "::saveGenericFile(): could not find load sub for mode '$mode': $!");
      return undef;
    }

  ##-- get fh
  my $fh = IO::File->new(">$filename");
  if (!$fh) {
    confess((ref($that)||$that) , "::saveGenericFile(): open failed for '$filename': $!");
    return undef;
  }
  binmode($fh,':utf8');

  my $rc = $sub->($that, $fh, %args);
  $fh->close();
  return $rc;
}


## $obj = $obj->saveGenericFh($fh,%args)
*saveFh = \&saveGenericFh;
sub saveGenericFh {
  my ($that,$fh,%args) = @_;
  my $mode = defined($args{mode}) ? $args{mode} : 'DEFAULT';
  delete($args{mode});

  ##-- get load sub
  if (!defined($sub=$that->ioModes->{$mode}{saveFh})
      ||
      (!ref($sub) && !defined($sub=UNIVERSAL::can($that,$sub))))
    {
      confess((ref($that)||$that), "::saveGenericFh(): could not find save sub for mode '$mode': $!");
      return undef;
    }

  my $rc = $sub->($that, $fh, %args);
  $fh->close();
  return $rc;
}


1;


##======================================================================
## Docs
=pod

=head1 NAME

WhoDat::Object - base class for WhoDat classes

=head1 SYNOPSIS

 use WhoDat::Object;

 ##-------------------------------------------------------------
 ## Derived Class: Definition
 ##-------------------------------------------------------------
 package MyObject;
 our @ISA = qw(WhoDat::Object);

 ##-- I/O: Native: Overrides
 sub saveNativeFh { ... }
 sub loadNativeFh { ... }

 ##-- I/O: Generic
 sub saveMyFh { ... }
 sub loadMyFh { ... }
 __PACKAGE__->registerIOMode('MyMode',{saveFh=>'saveMyFh',loadFh=>'loadMyFh'});
 __PACKAGE__->registerFileSuffix('.mine', 'MyMode');

 ##-------------------------------------------------------------
 ## Derived Class: Usage
 ##------------------------------------------------------------- 
 package main;

 $obj  = MyObject->new($arg=>$val,...);
 $obj2 = $obj->copy();

 $obj = $obj->loadFile("$filename",@args);           ##-- calls loadNativeFh()
 $obj = MyObject->loadFile("$filename",@args);       ##-- calls loadNativeFh()
 $obj = $obj->loadFile("$filename.mine",@args);      ##-- calls loadMyFh()

 $bool = $obj->saveFile($filename,@args);            ##-- calls saveNativeFh()
 $bool = $obj->saveFile("$filename.mine",@args);     ##-- calls saveMyFh()

=cut

##======================================================================
## Description
=pod

=head1 DESCRIPTION

WhoDat::Object is an abstract base class for WhoDat classes.
It offers methods for deep-object copying, as well as abstract and
extensible librarian functions.

=cut

##======================================================================
## METHODS
=pod

=head1 METHODS

=cut

##======================================================================
## METHODS: Constructors, etc.
=pod

=head2 Constructors, etc.

=over 4

=item $obj = $class_or_obj-E<gt>new(%args)

Default constructor: returns blessed hash reference initialized
to \%args.

=item $obj2 = $obj1->copy()

Deep-object copy, handles HASH-, ARRAY-, and SCALAR references.

=back

=cut

##======================================================================
## METHODS: I/O
=pod

=head2 I/O Methods

=over 4

=item $bool = $obj-E<gt>saveNativeFh($handle,@args)

Dummy method, should be overridden by child classes.

=item $bool = $obj-E<gt>saveNativeFile($filename,@args)

Wrapper for $obj-E<gt>saveNativeFh($fh,@args)

=item $bool = $obj-E<gt>saveFile($filename,%args)

Automagically determines file type for save.


=item $obj = $class_or_obj-E<gt>loadNativeFh($handle,@args)

Dummy method, should be overridden by child classes.

=item $obj = $class_or_obj-E<gt>loadNativeFile($filename,@args)

Wrapper for $obj-E<gt>loadNativeFh($fh,@args).

=item $bool = $class_or_obj-E<gt>loadFile($filename,%args)

Automagically determines file type for load.


=item $class-E<gt>registerIOMode($modename,\%details)

Registers a new I/O mode for objects of class $class.

=item $class->E<gt>registerFileSuffix($suffix,$modename)

Automatically selects mode "$modename" for load/save from/to
files with suffix "$suffix", for use with loadFile(), saveFile().

=back

=cut


##======================================================================
## Footer
=pod

=head1 ACKNOWLEDGEMENTS

perl by Larry Wall.

=head1 AUTHOR

Bryan Jurish E<lt>jurish@ling.uni-potsdam.deE<gt>

=head1 COPYRIGHT

Copyright (c) 2005, Bryan Jurish.  All rights reserved.

This package is free software.  You may redistribute it
and/or modify it under the same terms as Perl itself.

=head1 SEE ALSO

perl(1),
WhoDat(3perl).

=cut
