#!/usr/bin/perl -w

use lib qw(.);
use WhoDat;
use Getopt::Long;
use Pod::Usage;
use File::Basename qw(basename);

##======================================================================
## Globals
##======================================================================
our $lexfile  = undef;
our $nsents   = 1;
our $outfile  = '-';

our $progname = basename($0);

##======================================================================
## Options
##======================================================================
GetOptions(##-- General
	   'h|help' => \$help,
	   ##-- Model,
	   'l|lex|lexicon=s' => \$lexfile,
	   ##-- Generation
	   'o|out|output=s' => \$outfile,
	  );

pod2usage({
	   -exitval=>0,
	   -verbose=>0,
	  }) if ($help);
pod2usage({
	   -exitval=>0,
	   -verbose=>0,
	   -msg=>'You must specify a lexicon with the -lex option!',
	  }) if (!$lexfile);


if ($version) {
  print STDERR "$progname (WhoDat v$WhoDat::VERSION) by Bryan Jurish\n";
  exit 0 if ($version);
}


##======================================================================
## MAIN
##======================================================================

##-- open output file
open(OUT,">$outfile")
  or die("$progname: could not open output file '$outfile': $!");

##-- load lex file
our $lex = WhoDat::Lex->loadFile($lexfile)
  or die("$progname: could not load lexicon file '$lexfile': $!");

##-- generate
while (<>) {
  next if (/^%%/);
  if (/^\s*$/) {
    print OUT "\n";
    next;
  }
  chomp;

  $word = $lex->next($_);
  print OUT $word, ' ';
}

close OUT;

__END__

##======================================================================
## Pods
##======================================================================


=head1 NAME

whodat-genlex.perl - apply a WhoDat lexicon to a syntactic template

=head1 SYNOPSIS

 whodat-gensym.perl [OPTIONS] [TEMPLATE(s)...]

 General Options:
   -help
   -version

 Model Options:
   -lexicon LEXICON_FILE

 I/O Options:
   -output OUTPUT_FILE

=cut

###############################################################
## ARGUMENTS
###############################################################
=pod

=head1 ARGUMENTS

=over 4

=item TEMPLATE(s)

Specifies syntactic template files as written by the
whodat-gensyn.perl program.
Default is to read syntactic templates from stdin.

=back

=cut


###############################################################
## OPTIONS
###############################################################
=pod

=head1 OPTIONS

=cut

###############################################################
# General Options
###############################################################
=pod

=head2 General Options

=over 4

=item -help

Display a brief help message and exit.

=item -version

Display version information and exit.

=back

=cut


###############################################################
# Model Options
###############################################################
=pod

=head2 Model Options

=over 4

=item -lexicon LEXICON_FILE

Use lexical model file LEXICON_FILE.
LEXICON_FILE should be an lexical parameter file (*.lex)
as generated by the mootrain(1) program.
This is a required parameter.

=back

=cut

###############################################################
# I/O Options
###############################################################
=pod

=head2 I/O Options

=over 4

=item -output OUTPUT_FILE

Write output to OUTPUT_FILE.
Default='-' (write to stdout).

=back

=cut


###############################################################
# Bugs and Limitations
###############################################################
=pod

=head1 BUGS AND LIMITATIONS

Probably many.

=cut


###############################################################
# Footer
###############################################################
=pod

=head1 ACKNOWLEDGEMENTS

Perl by Larry Wall.

=head1 AUTHOR

Bryan Jurish E<lt>moocow@ling.uni-potsdam.deE<gt>

=head1 SEE ALSO

perl(1).

=cut
